module KXML
	class Document
		attr_accessor :encoding
		attr_accessor :version
		attr_accessor :root
		
		def initialize(data)
			case data
			when KXML::Node
				@root = data
			when String
				if data =~ /^[<>\/]/
					parse_from_string(data)
				else
					@root = Node.new(data)
				end
			when File
				parse_from_file(data)
			else
				raise "Unsupport Param Type : #{data.class}"
			end
			yield @root if block_given?
		end
		
		def save_to(file_path)
			File.open(file_path,"w+"){|f|
				f.write to_s
			}
		end
		
		def to_s(format = Format::PRETTY)
			return @root.to_s(format)
		end
		
		private
		
		def parse_from_string(data)
			pre_parse()
			push_data(data)
			post_parse()
		end
		
		def parse_from_file(file)
			pre_parse()
			line = file.gets
			while(line != nil)
				push_data(line)
				line = file.gets
			end
			post_parse()
		end
		
		def pre_parse
			require File.expand_path('../xml_parser.rb',__FILE__)
			@__parser = SAXParser.new
			@__tag_stack = []
			@__parser.handler=->(param){
				case param[:type]
				when SAXParser::DOCUMENT_STATEMENT
					@encoding = param[:attributes]['encoding']
					@version = param[:attributes]['version']
				when SAXParser::START_TAG
					node = Node.new(param[:name])
					node.attributes.merge!(param[:attributes]) if param[:attributes] != nil
					if @root == nil
						@root = node
					else
						@__tag_stack.last << node
					end
					@__tag_stack << node
				when SAXParser::END_TAG
					raise "Tag mismatch : <#{@__tag_stack.last.name}> - </#{param[:name]}>" if @__tag_stack.last ==nil || @__tag_stack.last.name != param[:name]
					@__tag_stack.pop
				when SAXParser::SELF_CLOSE_TAG
					node = Node.new(param[:name])
					node.attributes.merge!(param[:attributes]) if param[:attributes] != nil
					if @root == nil
						@root = node
					else
						@__tag_stack.last << node
					end
				when SAXParser::COMMENTS
					#ignore
				when SAXParser::CONTENT
					next if @__tag_stack.last == nil
					@__tag_stack.last.content = '' if @__tag_stack.last.content == nil
					@__tag_stack.last.content += param[:name].gsub(/\s/,"")
				end
			}
		end
		
		def push_data(data)
			@__parser << data
		end
		
		def post_parse
			@__parser.finish
			
			@__parser = nil
			@__tag_stack = nil
		end
	end
	
	class Node
		# @_children
		# @_content
		# @_name
		# @_attributes
		# @_parent
		def initialize(data)
			@_children = []
			@_attributes = {}
			case data
			when String
				if data =~ /^[<>\/]/
					parse_from_string(data)
				else
					@_name = data
				end
			when File
				parse_from_file(data)
			else
				raise "Unsupport Param Type : #{data.class}"
			end
			yield self if block_given?
		end
		
		def add_child(node)
			@_children << node
			node.parent = self
			return self
		end
		
		def <<(data)
			case data
			when Hash
				data.each{|k,v|
					k = k.to_s if k.is_a? Symbol
					@_attributes[k] = v
				}
			when KXML::Node
				add_child(data)
			when String
				if data =~ /^[<>\/]/
					self << Node.new(data)
				else
					append_content(data)
				end
			when Symbol
				append_content(data.to_s)
			when Array
				data.each{|e|
					self << e
				}
			else
				raise "method << not support this type : #{data.class}"
			end
			return self
		end
		
		def append_content(data)
			@_content = "" if @_content == nil
			@_content += data
		end
		
		def next_node
			p = parent
			return nil if p==nil
			index = -1
			p.children.each_with_index{|c,i|
				index = i if c==self
			}
			return nil if index == -1
			return p.children[index+1] if p.children[index+1]!=nil
			return nil
		end
		
		def last_node
			p = parent
			return nil if p==nil
			index = -1
			p.children.each_with_index{|c,i|
				index = i if c==self
			}
			return nil if index == -1 or index == 0
			return p.children[index-1] if p.children[index-1]!=nil
			return nil
		end
		
		def to_ary
			return [] << to_s
		end
		
		def to_s(format = Format::PRETTY)
			case format
			when Format::MINIMUM
				attributes_str = ""
				@_attributes.each{|key,value|
					attributes_str += " #{key}=\"#{value}\""
				}
				content = ""
				@_children.each{|child|
					content += child.to_s
				}
				content += @_content if @_content != nil
				return "<#{@_name}#{attributes_str}>#{content}</#{@_name}>"
			when Format::PRETTY
				return _to_s_pretty
			else
				raise "Unsupport Format : #{format}"
			end
		end
		
		def _to_s_pretty(deepth = 0)
			attributes_str = ""
			indent = "\t" * deepth
			wrap = "\n"
			@_attributes.each{|key,value|
				attributes_str += " #{key}=\"#{value}\""
			}
			content = ""
			@_children.each{|child|
				content += wrap
				content += child._to_s_pretty(deepth+1)
			}
			content += wrap if @_children.size > 0
			content += @_content if @_content != nil
			return "#{indent}<#{@_name}#{attributes_str}>#{content}#{@_children.size > 0 ? indent : ""}</#{@_name}>"
		end
		
		# getter setter
		
		def parent
			return @_parent
		end
		
		def [](key)
			return @_attributes[key.to_s]
		end
		
		def []=(key,value)
			@_attributes[key.to_s] = value
		end
		
		def attributes
			return @_attributes
		end
		
		def content
			return @_content
		end
		
		def content=(value)
			@_content = value
		end
		
		def name
			return @_name
		end
		
		def name=(value)
			@_name = value
		end
		
		def children
			return @_children
		end
		
		def call(*arg)
			if arg[1].class == Array then
				find_children_by_name(arg[0])
			else
				return find_child_by_name(arg[0])
			end
		end
		
		protected
		
		def parent=(p)
			@_parent = p
		end
		
		private
		
		def parse_from_string(data)
			document = Document.new(data)
			@_name = document.root.name
			@_children = document.root.children
			@_content = document.root.content
			@_attributes = document.root.attributes
		end
		
		def parse_from_file(file)
			document = Document.new(file)
			@_name = document.root.name
			@_children = document.root.children
			@_content = document.root.content
			@_attributes = document.root.attributes
		end
		
		def find_child_by_name(name)
			@_children.each{|child|
				return child if name.to_s == child.name.to_s
			}
			return nil
		end
		
		def find_children_by_name(name)
			ret = []
			@_children.each{|child|
				ret << child if name.to_s == child.name.to_s
			}
			return ret
		end
		
		
		def method_missing(method_name,*args,&block)
			if respond_to? method_name then
				super
			else
				if args[0].class == Array then
					return find_children_by_name(method_name)
				else
					return find_child_by_name(method_name)
				end
			end
		end
		
	end
	
	class Format
		MINIMUM = 'minimum'
		PRETTY	= 'pretty'
	end
	
end

